#include "global.h"
#include "io.h"
#include "z80.h"
#include "vdp.h"
#include "draw.h"
#include "crystal.h"
#include "state.h"
#include "settings.h"
#include "resource.h"
#include "main.h"
#include "msx.h"
#include "tool.h"
#include "reverse.h"

/*

Tested on my MSX1 (Canon V-20) VDP: TMS9929ANL, DDU delta 44926

Known models on MSXes and their differences:
* Texas Instruments TMS9918/9928/9929 (the original):
 - mixed mode
 - sprite cloning glitch
 - 4KB mode mixes up address lines
* Texas Instruments TMS9118/9128/9129:
 - mixed mode
 - sprite cloning glitch (officially documented)
 - 4KB mode no effect (officially documented)
* Toshiba T6950/T7937A: (not tested well yet)
 - no mixed mode
 - 5th sprite detection on screen 2, only works on upper 1/3rd part
 - no M3 pattern table mirroring
* Yamaha: (not tested well yet)
 - no sprite cloning glitch
 - mixed mode?

Small scrolling glitches (also happens on real MSX1, lack of bandwidth probably):
 Alpha Roid, Circus Charlie, Pippols, Teki-Paki, ...

Unclear:
* For pixel-accurate timing, need exact timing and behaviour of VDP memory fetches
 (current scanline implementation causes bugs in advanced demos, eg. the ones by dvik/joyrex. bugs in old "Waves" are normal)

* Asynchronous statusregister reads (also mentioned in official technical documentation)

* pitpan: "SONY MSX1 models do not support this "hybrid" mode: confirmed (personally) on SONY HB-10, HB-20, HB-101, HB-75."
 (about undocumented screenmodes (on a Toshiba VDP), emulated here)

* TMS99xx 4KB/16KB causing RAM degrading from lack of refresh

* Mid-screen statusregister reads (I just emulate the sprite number, rest is too unclear):
 
 The MSX can send a request to this register, to prevent its lower bits from updating, by setting bit 6.
 This request is only honored if bit 6 and 7 are clear. Afterwards, these bits will be "stuck", until the
 status register is read. These rules apply to all screen modes.
 
 Screen 1,2,3:
 Reading this register returns the current value of the internal sprite counter. It's possible to use this
 as a per-scanline horizontal counter, though some reads will return erroneous values (possibly due to reading
 it exactly the same time it's being updated). Of course, if the VDP sets the overflow bit, bits 0-4 will stay stuck.
 The sprite counter keeps continuing though, and if you reread it (twice) inside vblank, it will point to the
 last handled sprite number. Some simple examples (overflow or no overflow doesn't matter):
 sprite 16 Y value of 208 --> 16
 sprites 0,1,2,3,4 on scanline 191 --> 4
 no sprites on scanline 191 --> 31
 
 So, those bits don't just get reset to 00 or 31 after the status register is read.
 
 Screen 0:
 The collision bit is never set. Bits 6,4,3,2,1,0 represent an undocumented VDP register somehow related
 to the pattern name table, or maybe just open-bus, I haven't figured out the pattern. The on-screen test numbers
 (contents of the nametable) influencing themselves didn't help much :P .. values were not random though.
 
 Blank (reg 1 bit 6 clear):
 If the blank bit is cleared in mid-screen, the result will be the same (or similar, don't know) to those of screen 0.
 If it's cleared before vblank ends, it gives an interesting result (reads separated by a few cycles):
 scanline -1: (VDP activates: sprite evaluation starts here) garbage like screen 0
 scanline 0: $11, $11, $11, $11, $11, $51
 scanline 1: $01, $01, $01, $01, $01, $41
 scanline 2: $11, $11, $11, $11, $11, $51
 etc.
 Reading from this seems stable, unlike reading from the active sprite counter. Also, it seems that bit 6 doesn't
 get reset by reading this register (maybe it gets updated very fast).
 
 All these internal registers/counters/whatevers are inactive during vblank.

* Sprite cloning (different behaviour on MSXes):
 
 This effect only happens when reg 3 bit 6 and reg 4 bit 1 are 0 (other bits: don't care), in Screen 2.
 The internal sprite comparator seems to mess up. It will compare spriteline Y positions with the current line,
 and additionally with the current line OR 128, causing sprites in the 0-127 area to be cloned 128 pixels down
 (or on line 0 if Y=127) [[*edit* of course, clonesprites with a result Y of 191-254 are not handled]].
 Any normal sprite below this area will not be cloned. A normal sprite with Y=80 will not be interpreted
 as the end of the table (80+128=208). Other sprite functions remain normal: collision and overflow can happen
 between cloned and normal sprites. One thing to note is that the first 8 sprites are unaffected: they will
 not be cloned, the maximum number of sprites that can be shown with this is 56.
 
 Now for something very strange: with time (maybe due to temperature) this cloning effect gradually deteriorates
 (cloned sprites start to flicker, and it sometimes fetches a wrong X coordinate). It takes a few minutes for it
 to completely go away, and will only come back if I turn off my MSX1 for a while.
 
 FOLLOWUP: Of all the MSXes it was tested on, the deteriorating only happens on mine, with the colourtable set
 at 0 making it even worse. Not counting this weird effect, if M3 is set, sprites 8-31 Y position(s) is influenced
 by the least 2 bits of the pattern generator table offset, the colour table offset doesn't matter. The sprite
 comparator will detect a sprite at attr_y, attr_y OR (~reg4 << 6 AND 64), attr_y OR (~reg4 << 6 AND 128), and
 attr_y OR (~reg4 << 6 AND 192) when bits 0 and 1 are 0. It's as simple as that. I assume the reason for the first
 8 sprites not being affected is due to them being preprocessed in hblank. This glitch is briefly mentioned in the
 official TMS91xx programming manual btw, meaning that it's not undocumented.
 
 Internally, the description above seems to be the result from the rowcounter being masked, similar to the background
 patterns, though different for the first line.
 flyguille confirmed the deteriorating to be temperature related, using a blow dryer.

*/

#define RAMINIT	0	/* value RAM is filled with on power on. $ff on my Philips VG-8235, 00 on even and $ff on uneven address on my Canon V-20, Universe Unknown expects it to be 0 */

static int vdp_chiptype=0;
static int viewbg;
static int viewspr;
static int unlim;	/* will cause glitches in games that use overflow to hide sprite lines (eg. Antarctic Adventure, Athletic Land) */

static u8 ram[0x4000];
static int vblank=FALSE;

#define UPLOAD_MAX 0x100
static int upload_pending=FALSE;
static int upload_count=0;
static int upload_address[UPLOAD_MAX];
static int upload_size[UPLOAD_MAX];
static u8* upload_data[UPLOAD_MAX];

/* regs */
static int regs[8];	/* 8 control registers */
static int bgc=0;	/* bg colour at vblank */
static int status=0;	/* status register */
static int statuslow=0;

static int latch=0;	/* 1st/2nd write */
static int address=0;
static int read=0;	/* read ahead buffer */

static const char* vdp_chiptype_name[VDP_CHIPTYPE_MAX]={
	"TMS9118",	"TMS9129",	"TMS9918",	"TMS9929"
};
static const int vdp_chiptype_uid[VDP_CHIPTYPE_MAX]={
	0,		1,		2,		3
};

const char* vdp_get_chiptype_name(u32 i) { if (i>=VDP_CHIPTYPE_MAX) return NULL; else return vdp_chiptype_name[i]; }
int vdp_get_chiptype(void) { return vdp_chiptype; }
void vdp_set_chiptype(u32 i) { if (i>=VDP_CHIPTYPE_MAX) i=0; vdp_chiptype=i; }
void vdp_set_chiptype_vf(int i) { vdp_chiptype=(vdp_chiptype&2)|(i!=0); }
int vdp_get_chiptype_vf(void) { return vdp_chiptype&1; }

int vdp_get_chiptype_uid(u32 type)
{
	if (type>=VDP_CHIPTYPE_MAX) type=0;
	return vdp_chiptype_uid[type];
}

int vdp_get_uid_chiptype(u32 id)
{
	int type;
	
	for (type=0;type<VDP_CHIPTYPE_MAX;type++) {
		if (vdp_chiptype_uid[type]==id) return type;
	}
	
	return 0;
}

int __fastcall vdp_get_reg(u32 r) { return regs[r&7]; }
int __fastcall vdp_get_status(void) { return status; }
int __fastcall vdp_get_statuslow(void) { return statuslow; }
u8* __fastcall vdp_get_ram(void) { return ram; }

int vdp_get_bgc(void) { return bgc; }
void vdp_set_bgc(int c) { bgc=c; }
int vdp_get_bg_enabled(void) { return viewbg==0; }
void vdp_set_bg_enabled(int i) { main_menu_check(IDM_LAYERB,i); viewbg=i?0:8; }
int vdp_get_spr_enabled(void) { return viewspr!=0; }
void vdp_set_spr_enabled(int i) { main_menu_check(IDM_LAYERS,i); viewspr=i?0xff:0; }
int vdp_get_spr_unlim(void) { return unlim; }
void vdp_set_spr_unlim(int i) { main_menu_check(IDM_LAYERUNL,i); unlim=i; }

void vdp_init(void)
{
	char* c=NULL;
	int i=8; while (i--) regs[i]=0;
	for (i=0;i<UPLOAD_MAX;i++) upload_data[i]=NULL;
	
	i=settings_get_yesnoauto(SETTINGS_LAYERB);	if (i!=FALSE&&i!=TRUE) i=TRUE;	vdp_set_bg_enabled(i);
	i=settings_get_yesnoauto(SETTINGS_LAYERS);	if (i!=FALSE&&i!=TRUE) i=TRUE;	vdp_set_spr_enabled(i);
	i=settings_get_yesnoauto(SETTINGS_LAYERUNL);	if (i!=FALSE&&i!=TRUE) i=FALSE;	vdp_set_spr_unlim(i);
	
	i=VDP_CHIPTYPE_TMS9129;
	SETTINGS_GET_STRING(settings_info(SETTINGS_VDPCHIP),&c);
	if (c&&strlen(c)) {
		for (i=0;i<VDP_CHIPTYPE_MAX;i++) if (stricmp(c,vdp_get_chiptype_name(i))==0) break;
		if (i==VDP_CHIPTYPE_MAX) i=VDP_CHIPTYPE_TMS9129;
	}
	MEM_CLEAN(c);
	vdp_chiptype=i;
}

void vdp_clean(void)
{
	int i;
	for (i=0;i<UPLOAD_MAX;i++) { MEM_CLEAN(upload_data[i]); }
}

int vdp_upload(u32 a,u8* data,u32 size)
{
	if ((a&(VDP_UPLOAD_REG-1))>0x3fff||data==NULL||size==0||(a&VDP_UPLOAD_REG&&size!=1)||size>(0x4000-a)||upload_pending&2) return FALSE;
	
	if (msx_get_paused()) {
		/* only copy/invalidate when it differs */
		if (a&VDP_UPLOAD_REG) {
			/* reg */
			if (data[0]!=regs[a&7]) {
				regs[a&7]=data[0];
				tool_copy_locals(); reverse_invalidate();
			}
		}
		else if (memcmp(ram+a,data,size)) {
			/* ram */
			memcpy(ram+a,data,size);
			tool_copy_locals(); reverse_invalidate();
		}
	}
	else {
		if (upload_count==UPLOAD_MAX) return FALSE;
		
		upload_pending|=1;
		MEM_CREATE_N(upload_data[upload_count],size);
		memcpy(upload_data[upload_count],data,size);
		upload_size[upload_count]=size;
		upload_address[upload_count]=a;
		upload_count++;
	}
	
	return TRUE;
}

void vdp_poweron(void)
{
	int i=8;
	while (i--) regs[i]=0;
	status=statuslow=latch=address=read=0;
	
	memset(ram,RAMINIT,0x4000);
}

void vdp_reset(void)
{
	regs[0]=regs[1]=0;
	status=statuslow;
}

void vdp_new_frame(void)
{
	/* upload data */
	if (upload_pending) {
		int i;
		int c=FALSE;
		
		upload_pending|=2;
		
		for (i=0;i<UPLOAD_MAX;i++) {
			if (upload_data[i]==NULL) break;
			
			if (upload_address[i]&VDP_UPLOAD_REG) {
				/* reg */
				if (upload_data[i][0]!=regs[upload_address[i]&7]) {
					regs[upload_address[i]&7]=upload_data[i][0];
					c=TRUE;
				}
			}
			else if (memcmp(ram+upload_address[i],upload_data[i],upload_size[i])) {
				/* ram */
				memcpy(ram+upload_address[i],upload_data[i],upload_size[i]);
				c=TRUE;
			}
			
			MEM_CLEAN(upload_data[i]);
		}
		
		if (c) reverse_invalidate();
		upload_count=0;
		upload_pending=FALSE;
	}
	
	vblank=FALSE;
}

void vdp_vblank(void)
{
	vblank=TRUE;
	bgc=regs[7]&0xf; /* remember background colour */
	
	status|=0x80;
	if (regs[1]&0x20) z80_set_interrupt();
}


/* data read */
u8 __fastcall vdp_read_data(void)
{
	u8 ret=read;
	read=ram[address];
	address=(address+1)&0x3fff;
	latch=0;
	return ret;
}

/* data write */
void __fastcall vdp_write_data(u8 v)
{
	read=ram[address]=v;
	address=(address+1)&0x3fff;
	latch=0;
}

/* address/reg write */
void __fastcall vdp_write_address(u8 v)
{
	if (latch) {
		address=(v<<8|(address&0xff))&0x3fff;
		
		/* reg */
		if (v&0x80) {
			
			/* reg 1 special cases */
			if ((v&7)==1) {
				
				/* interrupt */
				if (status&0x80) {
					if (address&0x20) z80_set_interrupt();
					else z80_ack_interrupt();
				}
				
				/* TMS99xx 4k/16k mode */
				if (vdp_chiptype&2&&(regs[1]^address)&0x80) {
					u8 buf[0x4000];
					int i=0x4000;
					
					if (address&0x80) while (i--) buf[i]=ram[(i&0x203f)|(i>>1&0xfc0)|(i<<6&0x1000)];
					else while (i--) buf[i]=ram[(i&0x203f)|(i<<1&0x1f80)|(i>>6&0x40)];
					
					memcpy(ram,buf,0x4000);
				}
			}
			
			regs[v&7]=address&0xff;
		}
		
		/* read */
		else if (~v&0x40) { read=ram[address]; address=(address+1)&0x3fff; }
	}
	else address=(address&0x3f00)|v;
	
	latch^=1;
}

/* status read */
u8 __fastcall vdp_read_status(void)
{
	u8 ret=status;
	latch=0;
	status=statuslow;
	
	/* (async vblank?) */
	#if 0
	if (*z80_get_cycles_ptr()<=crystal->vblank_trigger&&!vblank) printf("x");
	#endif
	
	z80_ack_interrupt();
	return ret;
}


/* render 1 scanline */
void __fastcall vdp_line(int line)
{
	u8* screen=draw_get_screen_ptr()+256*line;
	int mode=(regs[1]>>4&1)|(regs[1]>>1&4)|(regs[0]&2)|viewbg; /* bit 0: M1, bit 1: M3, bit 2: M2, bit 3: background disabled */
	int pn=regs[2]<<10&0x3fff;			/* pattern nametable */
	int pg=regs[4]<<11&0x3fff;			/* pattern generator table */
	int ct=regs[3]<<6;				/* colour table */
	int bd=regs[7]&0xf;				/* background colour */
	
	if (~regs[1]&0x40) { mode=5; statuslow=0x1f; }	/* blank screen */
	else if (mode&2) {
		pg=(line<<5&pg)|(pg&0x2000);
		if (mode&5) mode^=2;			/* mode combination */
	}
	
	switch (mode) {
		
		/* screen 1 */
		case 0: {
			int x,pd,p,cd;
			
			pn|=(line<<2&0xfe0);
			pg|=(line&7);
			
			/* 32 tiles, 8 pixels per tile */
			for (x=0;x<32;x++) {
				cd=ram[pn|x];
				pd=ram[cd<<3|pg];
				cd=ram[cd>>3|ct];
				
				#define UPD1() *screen++=(p=((pd<<=1)&0x100)?cd>>4:cd&0xf)?p:bd
				UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1();
				/* (screen 2 uses it too) */
			}
			
			break;
		}
		
		/* screen 0 */
		case 1: {
			int x,pd,tc=regs[7]>>4;
			
			/* left border */
			*screen++=bd; *screen++=bd; *screen++=bd; *screen++=bd;
			*screen++=bd; *screen++=bd; *screen++=bd; *screen++=bd;
			
			pn|=((line>>3)*40);
			pg|=(line&7);
			
			/* 40 tiles, 6 pixels per tile */
			for (x=0;x<40;x++) {
				pd=ram[ram[pn+x]<<3|pg];
				
				#define UPD0() *screen++=((pd<<=1)&0x100)?tc:bd;
				UPD0(); UPD0(); UPD0(); UPD0(); UPD0(); UPD0();
				#undef UPD0
			}
			
			/* right border */
			*screen++=bd; *screen++=bd; *screen++=bd; *screen++=bd;
			*screen++=bd; *screen++=bd; *screen++=bd; *screen++=bd;
			
			statuslow=0x1f;
			
			break;
		}
		
		/* screen 2 */
		case 2: {
			int x,pd,p,cd,ct_mask=ct|0x3f;
			
			pn|=(line<<2&0xfe0);
			pg|=(line&7);
			ct=(ct&0x2000)|(line&7)|(line<<5&0x1800);
			
			/* 32 tiles, 8 pixels per tile */
			for (x=0;x<32;x++) {
				cd=ram[pn|x]<<3;
				pd=ram[cd|pg];
				cd=ram[(cd|ct)&ct_mask];
				
				UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1(); UPD1();
				#undef UPD1
			}
			
			break;
		}
		
		/* screen 3 */
		case 4: {
			int x,pd,p;
			
			pn|=(line<<2&0xfe0);
			pg|=(line>>2&7);
			
			/* 32 tiles, 8 pixels per 'tile' */
			for (x=0;x<32;x++) {
				pd=ram[ram[pn|x]<<3|pg];
				
				/* left 4 pixels */
				p=pd>>4; p=p?p:bd;
				*screen++=p; *screen++=p; *screen++=p; *screen++=p;
				
				/* right 4 pixels */
				p=pd&0xf; p=p?p:bd;
				*screen++=p; *screen++=p; *screen++=p; *screen++=p;
			}
			
			break;
		}
		
		/* blank screen or bogus mode */
		default: {
			memset(screen,bd,256);
			screen+=256;
			
			break;
		}
	}
	
	/* sprites */
	if (~mode&1) {
		int size=8<<(regs[1]>>1&1),mag=regs[1]&1;
		int smask=(size&16)?0xfe0:0xfff;
		int smax=(size<<mag)-1; u8 yc;
		int y,f=0,o=0x20,clone=0x20,slmask=0xff,rowmask=0xff,dismask=0;
		u8* sa=ram+(regs[5]<<7&0x3fff);		/* sprite attribute table */
		u8* sg=ram+(regs[6]<<11&0x3fff);	/* sprite generator table */
		u8 sl[0x20+0x100+0x20];			/* sprite line */
		statuslow=0;
		
		line--; /* sprite preprocessing is actually done one line in advance */
		
		/* invalidate out of bounds area */
		memset(sl,0x80,0x20);
		memset(sl+0x20,0,0x100);
		memset(sl+0x120,0x80,0x20);
		
		/* init sprite cloning */
		if (regs[0]&2&&(regs[4]&3)!=3) {
			clone=7;
			if (line==-1) dismask=~regs[4]<<6&0xc0; /* combined mask */
			else rowmask=(regs[4]<<6&0xc0)|0x3f;
		}
		
		for (;;) {
			/* forced break */
			if ((y=*sa++)==208) break;
			
			/* found sprite */
			if ((yc=line-y)<=smax||(statuslow>clone&&(yc=(line&rowmask)-(y|dismask))<=smax)) {
				u8* slp; int cd,g,x=*sa+++32;
				int pd=sg[g=yc>>mag|(*sa++<<3&smask)];		/* pixel data */
				if (f++&4) {
					o|=((status>>1^0x40)&0x40);		/* overflow */
					if (unlim) {				/* unlimited sprites? continue */
						if (slmask&0x80) { o&=0x40; slmask=statuslow; }
					}
					else break;
				}
				if ((cd=*sa++)&0x80) x-=32; cd=(cd&0xf)|0x20;	/* colour data */
				
				slp=sl+x;
				
				/* normal 8 pixel update */
				#define UPDS() \
				if ((pd<<=1)&0x100) { \
					status|=(*slp&0x20&o);			/* collision */ \
					if (!(*slp&0x8f)) *slp=cd; \
				} slp++
				
				/* magnified (16 pixels) update */
				#define UPDSM() \
				if ((pd<<=1)&0x100) { \
					status|=((*slp|(*slp+1))&0x20&o); \
					if (!(*slp&0x8f)) *slp=cd; slp++; \
					if (!(*slp&0x8f)) *slp=cd; slp++; \
				} else slp+=2
				
				if (mag) { UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); }
				else { UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); }
				
				/* 16*16 */
				if (size&16) {
					pd=sg[g|16];
					
					if (mag) { UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); UPDSM(); }
					else { UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); UPDS(); }
				}
				
				#undef UPDS
				#undef UPDSM
			}
			else sa+=3;
			
			/* all sprites checked */
			if (statuslow==31) break;
			else statuslow++;
		}
		
		statuslow&=slmask;
		if (~status&0x40) status=(status&0xe0)|statuslow|(o&0x40);
		
		/* update screen */
		if (f&viewspr) {
			int x=256; u8* slp=sl+32;
			screen-=256;
			
			while (x--) {
				if (*slp&0xf) *screen=*slp&0xf;
				screen++; slp++;
			}
		}
	}
}


/* state		size
ram			0x4000
regs			8
vblank			1
bgc			1
status			1
statuslow		1
latch			1
address			2
read			1

==			0x4000+16
*/
#define STATE_VERSION	1
#define STATE_SIZE	(0x4000+16)

int __fastcall vdp_state_get_version(void)
{
	return STATE_VERSION;
}

int __fastcall vdp_state_get_size(void)
{
	return STATE_SIZE;
}

void __fastcall vdp_state_save(u8** s)
{
	STATE_SAVE_C(ram,0x4000);
	
	STATE_SAVE_1(regs[0]);	STATE_SAVE_1(regs[1]);	STATE_SAVE_1(regs[2]);	STATE_SAVE_1(regs[3]);
	STATE_SAVE_1(regs[4]);	STATE_SAVE_1(regs[5]);	STATE_SAVE_1(regs[6]);	STATE_SAVE_1(regs[7]);
	
	STATE_SAVE_1(vblank);
	STATE_SAVE_1(bgc);
	STATE_SAVE_1(status);	STATE_SAVE_1(statuslow);
	STATE_SAVE_1(latch);
	STATE_SAVE_2(address);
	STATE_SAVE_1(read);
}

void __fastcall vdp_state_load_cur(u8** s)
{
	STATE_LOAD_C(ram,0x4000);
	
	STATE_LOAD_1(regs[0]);	STATE_LOAD_1(regs[1]);	STATE_LOAD_1(regs[2]);	STATE_LOAD_1(regs[3]);
	STATE_LOAD_1(regs[4]);	STATE_LOAD_1(regs[5]);	STATE_LOAD_1(regs[6]);	STATE_LOAD_1(regs[7]);
	
	STATE_LOAD_1(vblank);
	STATE_LOAD_1(bgc);
	STATE_LOAD_1(status);	STATE_LOAD_1(statuslow);
	STATE_LOAD_1(latch);
	STATE_LOAD_2(address);
	STATE_LOAD_1(read);
}

int __fastcall vdp_state_load(int v,u8** s)
{
	switch (v) {
		case STATE_VERSION:
			vdp_state_load_cur(s);
			break;
		
		default: return FALSE;
	}
	
	return TRUE;
}
